"""
Function for visualising the `eop_lag`, `mop_lag` or `grading` metrics for a given
set of cross sections and extended categories.
```python
import macrosynergy.visuals as msv
...
msv.view_metrics(
    df,
    xcats="IR",
    cids=["USD", "EUR", "GBP"],
    metric='eop_lag'
)
...
```
"""

from typing import List, Optional, Tuple
import pandas as pd
import seaborn as sns

from macrosynergy.visuals import Heatmap


def view_metrics(
    df: pd.DataFrame,
    xcat: str,
    cids: Optional[List[str]] = None,
    start: Optional[str] = None,
    end: Optional[str] = None,
    freq: str = "M",
    agg: str = "mean",
    metric: str = "eop_lag",
    title: Optional[str] = None,
    figsize: Optional[Tuple[float]] = (14, None),
) -> None:
    """
    A function to visualise the `eop_lag`, `mop_lag` or `grading` metrics for a given
    JPMaQS dataset. It generates a heatmap, where the x-axis is the observation
    date, the y-axis is the ticker, and the colour is the lag value.

    :param <pd.Dataframe> df: standardized DataFrame with the necessary columns:
        'cid', 'xcat', 'real_date' and 'grading', 'eop_lag' or 'mop_lag'.
    :param str xcat: extended category whose lags are to be visualized.
    :param <List[str]> cids: cross sections to visualize. Default is all in DataFrame.
    :param <str> start: earliest date in ISO format. Default is earliest available.
    :param <str> end: latest date in ISO format. Default is latest available.
    :param <str> metric: name of metric to be visualized. Must be "eop_lag" (default)
        "mop_lag" or "grading".
    :param <str> freq: frequency of data. Must be one of "D", "W", "M", "Q", "A".
        Default is "M".
    :param <str> agg: aggregation method. Must be one of "mean" (default), "median",
        "min", "max", "first" or "last".
    :param <str> title: string of chart title; if none given default title is printed.
    :param <Tuple[float]> figsize: Tuple (w, h) of width and height of graph.
        Default is None, meaning it is set in accordance with df.

    :return: None

    :raises TypeError: if any of the inputs are of the wrong type.
    :raises ValueError: if any of the inputs are semantically incorrect.
    """

    heatmap = Heatmap(
        df=df,
        cids=cids,
        xcats=[xcat],
        metrics=[metric],
        start=start,
        end=end,
    )

    if title is None:
        title = f"Visualising {metric} for {xcat} from {heatmap.start} to {heatmap.end}"

    if not isinstance(figsize, tuple):
        if not isinstance(figsize, list) or len(figsize) != 2:
            raise TypeError("`figsize` must be a tuple or list of length 2")

    if not isinstance(figsize[0], (int, float)):
        raise ValueError("First element of `figsize` must be a float")

    if figsize[1] is None:
        figsize = (figsize[0], len(heatmap.cids) + 1)  # +1 to adjust for labels

    if not isinstance(figsize[1], (int, float)):
        raise ValueError("Second element of `figsize` must be a float")

    heatmap.plot_metric(
        x_axis_column="real_date",
        y_axis_column="cid",
        metric=metric,
        freq=freq,
        agg=agg,
        title=title,
        figsize=figsize,
        x_axis_label="Date",
        y_axis_label="Cross Sections",
        cmap=sns.color_palette("light:red", as_cmap=True),
        rotate_xticks=90,
    )


if __name__ == "__main__":
    from macrosynergy.management.simulate import make_test_df

    test_cids: List[str] = ["USD", "EUR", "GBP"]
    test_xcats: List[str] = ["FX", "IR"]
    dfE: pd.DataFrame = make_test_df(
        cids=test_cids, xcats=test_xcats, style="sharp-hill"
    )

    dfM: pd.DataFrame = make_test_df(
        cids=test_cids, xcats=test_xcats, style="four-bit-sine"
    )

    dfG: pd.DataFrame = make_test_df(cids=test_cids, xcats=test_xcats, style="sine")

    dfE.rename(columns={"value": "eop_lag"}, inplace=True)
    dfM.rename(columns={"value": "mop_lag"}, inplace=True)
    dfG.rename(columns={"value": "grading"}, inplace=True)
    mergeon = ["cid", "xcat", "real_date"]
    dfx: pd.DataFrame = pd.merge(pd.merge(dfE, dfM, on=mergeon), dfG, on=mergeon)

    view_metrics(
        df=dfx,
        xcat="FX",
    )
    view_metrics(
        df=dfx,
        xcat="IR",
        metric="mop_lag",
    )
    view_metrics(
        df=dfx,
        xcat="IR",
        metric="grading",
    )



from .lad_regressor import (
    LADRegressor,
)
from .weighted_regressors import (
    BaseWeightedRegressor,
    TimeWeightedRegressor,
    SignWeightedRegressor,
    WeightedLinearRegression,
    SignWeightedLinearRegression,
    TimeWeightedLinearRegression,
    WeightedLADRegressor,
    SignWeightedLADRegressor,
    TimeWeightedLADRegressor,
)
from .naive_predictor import (
    NaivePredictor,
)
from .regressor_systems import (
    BaseRegressionSystem,
    LADRegressionSystem,
    RidgeRegressionSystem,
    LinearRegressionSystem,
    CorrelationVolatilitySystem,
)

from .bootstrap import (
    ModifiedLinearRegression,
    BaseModifiedRegressor,
)

__all__ = [
    "LADRegressor",
    "NaivePredictor",
    "SignWeightedLADRegressor",
    "TimeWeightedLADRegressor",
    "SignWeightedLinearRegression",
    "TimeWeightedLinearRegression",
    "BaseWeightedRegressor",
    "WeightedLinearRegression",
    "WeightedLADRegressor",
    "TimeWeightedRegressor",
    "SignWeightedRegressor",
    "BaseRegressionSystem",
    "LADRegressionSystem",
    "RidgeRegressionSystem",
    "LinearRegressionSystem",
    "CorrelationVolatilitySystem",
    "ModifiedLinearRegression",
    "BaseModifiedRegressor",
]



from .utils.check_availability import (
    check_availability,
    visual_paneldates,
    check_enddates,
    check_startyears,
    missing_in_df,
    business_day_dif,
)

from .simulate import simulate_vintage_data, simulate_quantamental_data

from .simulate.simulate_vintage_data import VintageData
from .simulate.simulate_quantamental_data import make_qdf
from .utils import common_cids, update_df, reduce_df, categories_df, reduce_df_by_ticker, merge_categories
from . import utils, types, decorators, simulate, constants
from .validation import validate_and_reduce_qdf

from .utils.sparse import (
    create_delta_data,
    calculate_score_on_sparse_indicator,
    sparse_to_dense,
    temporal_aggregator_exponential,
    temporal_aggregator_period,
    temporal_aggregator_mean,
    InformationStateChanges,
)

__all__ = [
    # METHODS
    "reduce_df",
    "categories_df",
    "reduce_df_by_ticker",
    "visual_paneldates",
    "check_enddates",
    "check_startyears",
    "missing_in_df",
    "business_day_dif",
    "VintageData",
    "make_qdf",
    "common_cids",
    "update_df",
    # Modules/Subpackages
    "utils",
    "types",
    "decorators",
    "simulate",
    "constants",
    # Sparse Indicators
    "create_delta_data",
    "calculate_score_on_sparse_indicator",
    "sparse_to_dense",
    "temporal_aggregator_exponential",
    "temporal_aggregator_period",
    "temporal_aggregator_mean",
    "InformationStateChanges",
    # Module-as-methods
    "check_availability",
    "simulate_vintage_data",
    "simulate_quantamental_data",
    "validate_and_reduce_qdf",
    "merge_categories",
]



import numpy as np
import pandas as pd
import datetime

from collections import Counter, defaultdict
from abc import ABC

from typing import Union, Optional

class BasePanelBootstrap(ABC):
    def __init__(
        self,
        bootstrap_method: str = "panel",
        resample_ratio: Union[float, int] = 1,
        max_features: Optional[Union[str, int, float]] = None, # TODO
    ):
        """
        Base class to construct bootstrap datasets over a panel. 

        :param <str> bootstrap_method: Method to bootstrap the data. Current options are
            "panel", "period", "cross", "cross_per_period" and "period_per_cross".
            Default is "panel".
        :param <Union[float, int]> resample_ratio: The ratio of resampling units comprised
            in each bootstrap dataset. This is a fraction of the quantity of the panel
            component to be resampled. Default value is 1.
        :param <Optional[Union[str, int, float]]> max_features: The number of features to
            consider in each bootstrap dataset. This can be used to increase the 
            variation of the bootstrap datasets. Default is None and currently not
            implemented.
        """
        # Checks 
        self._check_boot_params(
            bootstrap_method=bootstrap_method,
            resample_ratio=resample_ratio,
            max_features=max_features,
        )

        # Set attributes
        self.bootstrap_method = bootstrap_method
        self.resample_ratio = resample_ratio
        self.max_features = max_features

    def create_bootstrap_dataset(
        self,
        X: pd.DataFrame,
        y: Union[pd.DataFrame, pd.Series],
    ):
        """
        Method to generate a bootstrap dataset based on a panel of features and a target.

        :param <pd.DataFrame> X: Pandas dataframe of input features.
        :param <Union[pd.DataFrame, pd.Series]> y: Pandas series or dataframe of targets
            associated with each sample in X.

        :return Tuple[pd.DataFrame, Union[pd.DataFrame, pd.Series]]: A tuple of the
            resampled feature matrix and target vector.
        """
        # Checks
        self._check_create_bootstrap_dataset_params(
            X,
            y,
        )

        # Store index information in numpy arrays
        index_array = np.array(X.index.tolist())
        cross_sections = index_array[:, 0]
        unique_cross_sections = np.unique(cross_sections)
        real_dates = index_array[:, 1]
        unique_real_dates = np.unique(real_dates)

        # Create a bootstrap dataset 
        if self.bootstrap_method == "panel":
            X_resampled, y_resampled = self._panel_bootstrap(
                X = X,
                y = y,
            )

        elif self.bootstrap_method == "period":
             X_resampled, y_resampled = self._period_bootstrap(
                X = X,
                y = y,
                unique_real_dates = unique_real_dates,
            )

        elif self.bootstrap_method == "cross":
            X_resampled, y_resampled = self._cross_bootstrap(
                X = X,
                y = y,
                unique_cross_sections = unique_cross_sections,
            )

        elif self.bootstrap_method == "cross_per_period":
            X_resampled, y_resampled = self._cross_per_period_bootstrap(
                X = X,
                y = y,
                unique_cross_sections = unique_cross_sections,
            )

        elif self.bootstrap_method == "period_per_cross":
            X_resampled, y_resampled = self._period_per_cross_bootstrap(
                X = X,
                y = y,
                unique_real_dates = unique_real_dates,
            )

        return X_resampled, y_resampled

    def _panel_bootstrap(
        self,
        X: pd.DataFrame,
        y: Union[pd.DataFrame, pd.Series],      
    ):
        """
        Method to generate a bootstrap dataset by resampling the entire panel.

        :param <pd.DataFrame> X: Pandas dataframe of input features.
        :param <Union[pd.DataFrame, pd.Series]> y: Pandas series or dataframe of targets
            associated with each sample in X.

        :return Tuple[pd.DataFrame, Union[pd.DataFrame, pd.Series]]: A tuple of the
            resampled feature matrix and target vector.
        """
        bootstrap_idx = np.random.choice(
            np.arange(X.shape[0]),
            size=int(np.ceil(self.resample_ratio * X.shape[0])),
            replace=True,
        )
        X_resampled = X.iloc[bootstrap_idx]
        y_resampled = y.iloc[bootstrap_idx]

        return X_resampled, y_resampled
    
    def _period_bootstrap(
        self,
        X: pd.DataFrame,
        y: Union[pd.DataFrame, pd.Series],
        unique_real_dates: np.ndarray,
    ):
        """
        Method to generate a bootstrap dataset by resampling periods in the panel.

        :param <pd.DataFrame> X: Pandas dataframe of input features.
        :param <Union[pd.DataFrame, pd.Series]> y: Pandas series or dataframe of targets
            associated with each sample in X.
        :param <np.ndarray[pd.Timestamp]> unique_real_dates: Numpy array of the unique
            dates in the panel.

        :return Tuple[pd.DataFrame, Union[pd.DataFrame, pd.Series]]: A tuple of the
            resampled feature matrix and target vector.
        """
        # Resample unique panel dates with replacement
        bootstrap_periods = np.random.choice(
            unique_real_dates,
            size=int(np.ceil(self.resample_ratio * len(unique_real_dates))),
            replace=True,
        )

        # Obtain a {count: [periods]} dictionary so that the new panel can be efficiently
        # constructed by looping through the counts instead of the periods. 
        period_counts = dict(Counter(bootstrap_periods))
        count_to_periods = defaultdict(list)
        for period, count in period_counts.items():
            count_to_periods[count].append(period)

        # For each count, extract the periods that have that count and tile then by count
        # to create the new panel.
        X_resampled = np.empty((0, X.shape[1]))
        y_resampled = np.empty(0)
        index_resampled = []
        for count, periods in count_to_periods.items():
            X_resampled = np.vstack(
                [
                    X_resampled,
                    np.tile(
                        X[X.index.get_level_values(1).isin(periods)].values,
                        (count, 1),
                    ),
                ]
            )
            y_resampled = np.append(
                y_resampled,
                np.tile(
                    y[y.index.get_level_values(1).isin(periods)].values, count
                ),
            )
            index_resampled.extend(
                X[X.index.get_level_values(1).isin(periods)].index.tolist() * count
            )

        # reconstruct index
        index_resampled = pd.MultiIndex.from_tuples(
            index_resampled, names=["cid", "real_date"]
        )

        # Convert resampled datasets to pandas dataframes
        X_resampled = pd.DataFrame(
            data = X_resampled,
            index = index_resampled,
            columns = X.columns,
        )
        y_resampled = pd.Series(
            data = y_resampled,
            index = index_resampled,
            name = y.name,
        )

        return X_resampled, y_resampled
    
    def _cross_bootstrap(
        self,
        X: pd.DataFrame,
        y: Union[pd.DataFrame, pd.Series],
        unique_cross_sections: np.ndarray,
    ):
        """
        Method to generate a bootstrap dataset by resampling cross-sections in the panel.

        :param <pd.DataFrame> X: Pandas dataframe of input features.
        :param <Union[pd.DataFrame, pd.Series]> y: Pandas series or dataframe of targets
            associated with each sample in X.
        :param <np.ndarray[pd.Timestamp]> unique_real_dates: Numpy array of the unique
            dates in the panel.

        :return Tuple[pd.DataFrame, Union[pd.DataFrame, pd.Series]]: A tuple of the
            resampled feature matrix and target vector.
        """
        # Resample unique panel cross-sections with replacement
        bootstrap_cross_sections = np.random.choice(
            unique_cross_sections,
            size=int(np.ceil(len(unique_cross_sections) * self.resample_ratio)),
            replace=True,
        )

        # Obtain a {count: [cross_sections]} dictionary so that the new panel can be efficiently
        # constructed by looping through the counts instead of the periods. 
        cross_section_counts = dict(Counter(bootstrap_cross_sections))
        count_to_cross_sections = defaultdict(list)
        for cross_section, count in cross_section_counts.items():
            count_to_cross_sections[count].append(cross_section)

        # For each count, tile the observations within the cross-sections with that count
        X_resampled = np.empty((0, X.shape[1]))
        y_resampled = np.empty(0)
        index_resampled = []
        for count, cross_sections in count_to_cross_sections.items():
            X_resampled = np.vstack(
                [
                    X_resampled,
                    np.tile(
                        X[X.index.get_level_values(0).isin(cross_sections)].values,
                        (count, 1),
                    ),
                ]
            )
            y_resampled = np.append(
                y_resampled,
                np.tile(
                    y[y.index.get_level_values(0).isin(cross_sections)].values, count
                ),
            )
            index_resampled.extend(
                X[X.index.get_level_values(0).isin(cross_sections)].index.tolist() * count
            )

        # reconstruct index
        index_resampled = pd.MultiIndex.from_tuples(
            index_resampled, names=["cid", "real_date"]
        )

        # Convert resampled datasets to pandas dataframes
        X_resampled = pd.DataFrame(
            data = X_resampled,
            index = index_resampled,
            columns = X.columns,
        )
        y_resampled = pd.Series(
            data = y_resampled,
            index = index_resampled,
            name = y.name,
        )

        return X_resampled, y_resampled
    
    def _cross_per_period_bootstrap(
        self,
        X: pd.DataFrame,
        y: Union[pd.DataFrame, pd.Series],
        unique_cross_sections: np.ndarray,      
    ):
        """
        Method to generate a bootstrap dataset by resampling cross-sections within each
        period in the panel.

        :param <pd.DataFrame> X: Pandas dataframe of input features.
        :param <Union[pd.DataFrame, pd.Series]> y: Pandas series or dataframe of targets
            associated with each sample in X.
        :param <np.ndarray[str]> unique_cross_sections: Numpy array of the unique
            cross-sections in the panel.

        :return Tuple[pd.DataFrame, Union[pd.DataFrame, pd.Series]]: A tuple of the
            resampled feature matrix and target vector.
        """
        n_resample = int(np.ceil(len(unique_cross_sections) * self.resample_ratio))
        X_resampled = X.groupby(level=1).sample(replace=True, n=n_resample)
        y_resampled = y.loc[X_resampled.index]

        return X_resampled, y_resampled
    
    def _period_per_cross_bootstrap(
        self,
        X: pd.DataFrame,
        y: Union[pd.DataFrame, pd.Series],
        unique_real_dates: np.ndarray,      
    ):
        """
        Method to generate a bootstrap dataset by resampling periods within each
        cross-section in the panel.

        :param <pd.DataFrame> X: Pandas dataframe of input features.
        :param <Union[pd.DataFrame, pd.Series]> y: Pandas series or dataframe of targets
            associated with each sample in X.
        :param <np.ndarray[pd.Timestamp]> unique_real_dates: Numpy array of the unique
            dates in the panel.

        :return Tuple[pd.DataFrame, Union[pd.DataFrame, pd.Series]]: A tuple of the
            resampled feature matrix and target vector.
        """
        n_resample = int(np.ceil(len(unique_real_dates) * self.resample_ratio))
        X_resampled = X.groupby(level=0).sample(replace=True, n=n_resample)
        y_resampled = y.loc[X_resampled.index]

        return X_resampled, y_resampled

    def _check_create_bootstrap_dataset_params(
        self,
        X: pd.DataFrame,
        y: Union[pd.DataFrame, pd.Series],        
    ):
        """
        Method to check the validity of the input parameters for create_bootstrap_dataset.

        :param <pd.DataFrame> X: Pandas dataframe of input features.
        :param <Union[pd.DataFrame, pd.Series]> y: Pandas series or dataframe of targets
        """
        # Checks
        if not isinstance(X, pd.DataFrame):
            raise TypeError(
                "Input feature matrix must be a pandas dataframe. "
                "If used as part of an sklearn pipeline, ensure that previous steps "
                "return a pandas dataframe."
            )
        if not (isinstance(y, pd.Series) or isinstance(y, pd.DataFrame)):
            raise TypeError(
                "Target vector must be a pandas series or dataframe. "
                "If used as part of an sklearn pipeline, ensure that previous steps "
                "return a pandas series or dataframe."
            )
        if isinstance(y, pd.DataFrame) and y.shape[1] != 1:
            raise ValueError(
                "The target dataframe must have only one column. If used as part of "
                "an sklearn pipeline, ensure that previous steps return a pandas "
                "series or dataframe."
            )

        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("X must be multi-indexed.")
        if not isinstance(y.index, pd.MultiIndex):
            raise ValueError("y must be multi-indexed.")
        if not isinstance(X.index.get_level_values(1)[0], datetime.date):
            raise TypeError("The inner index of X must be datetime.date.")
        if not isinstance(y.index.get_level_values(1)[0], datetime.date):
            raise TypeError("The inner index of y must be datetime.date.")
        if not X.index.equals(y.index):
            raise ValueError(
                "The indices of the input dataframe X and the output dataframe y don't "
                "match."
            )
        
    def _check_boot_params(
        self,
        bootstrap_method: str,
        resample_ratio: Union[float, int],
        max_features: Optional[Union[str, int, float]],
    ):
        """
        Private method to check the class initialization parameters.

        :param <str> bootstrap_method: Method to bootstrap the data. Current options are
            "panel", "period", "cross", "cross_per_period" and "period_per_cross".
        :param <Union[float, int]> resample_ratio: The ratio of resampling units comprised
            in each bootstrap dataset. This is a fraction of the quantity of the panel
            component to be resampled.
        :param <Optional[Union[str, int, float]]> max_features: The number of features to
            consider in each bootstrap dataset. This can be used to increase the 
            variation of the bootstrap datasets.

        :return None
        """
        # bootstrap_method
        if not isinstance(bootstrap_method, str):
            raise TypeError(
                f"bootstrap_method must be a string. Got {type(bootstrap_method)}."
            )
        if bootstrap_method not in [
            "panel",
            "period",
            "cross",
            "cross_per_period",
            "period_per_cross",
        ]:
            raise ValueError(
                f"bootstrap_method must be one of 'panel', 'period', 'cross',"
                " 'cross_per_period', 'period_per_cross'. Got {bootstrap_method}."
            )

        # resample_ratio
        if not isinstance(resample_ratio, (int, float)):
            raise TypeError(
                f"resample_ratio must be an integer or a float. Got {type(resample_ratio)}."
            )
        if resample_ratio <= 0:
            raise ValueError("resample_ratio must be greater than 0.")
        if resample_ratio > 1:
            raise ValueError("resample_ratio must be less than or equal to 1.")
        
        # max_features
        if max_features is not None:
            raise NotImplementedError("max_features is not implemented yet.")
        

if __name__ == "__main__":
    from macrosynergy.management.simulate import make_qdf
    import macrosynergy.management as msm

    # Simulate an unbalanced panel, multiindexed by cross-section and date
    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]
    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2002-01-01", "2020-12-31", 0, 1]
    df_cids.loc["CAD"] = ["2003-01-01", "2020-12-31", 0, 1]
    df_cids.loc["GBP"] = ["2000-01-01", "2020-12-31", 0, 1]
    df_cids.loc["USD"] = ["2000-01-01", "2020-12-31", 0, 1]

    df_xcats = pd.DataFrame(index=xcats, columns=cols)
    df_xcats.loc["XR"] = ["2000-01-01", "2020-12-31", 0, 1, 0, 3]
    df_xcats.loc["CRY"] = ["2000-01-01", "2020-12-31", 0, 1, 0, 0]
    df_xcats.loc["GROWTH"] = ["2000-01-01", "2020-12-31", 0, 1, -0.9, 0]
    df_xcats.loc["INFL"] = ["2000-01-01", "2020-12-31", 0, 1, 0.8, 0]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    dfd["grading"] = np.ones(dfd.shape[0])
    black = {"GBP": ["2009-01-01", "2012-06-30"], "CAD": ["2018-01-01", "2100-01-01"]}
    dfd = msm.reduce_df(df=dfd, cids=cids, xcats=xcats, blacklist=black)

    dfd = dfd.pivot(index=["cid", "real_date"], columns="xcat", values="value")
    X = dfd.drop(columns=["XR"])
    y = dfd["XR"]

    bootstrap_methods = [
        "panel",
        "period",
        "cross",
        "cross_per_period",
        "period_per_cross",
    ]

    for method in bootstrap_methods:
        # Initialize the BasePanelBootstrap class
        bpb = BasePanelBootstrap(
            bootstrap_method=method,
            resample_ratio=0.8,
        )
        print(bpb.create_bootstrap_dataset(X, y))


"""
Run Granger Causality Test on a standardized quantamental dataframe.

**NOTE: This module is under development, and is not yet ready for production use.**
"""

import warnings
from typing import Any, Dict, List, Optional, Union

import pandas as pd
import statsmodels
from packaging import version
from statsmodels.tsa.stattools import grangercausalitytests

from macrosynergy.management.simulate import make_test_df
from macrosynergy.management.types import QuantamentalDataFrame
from macrosynergy.management.utils import (
    downsample_df_on_real_date,
    is_valid_iso_date,
    qdf_to_ticker_df,
    reduce_df_by_ticker,
)

import logging

logger = logging.getLogger(__name__)


def _statsmodels_compatibility_wrapper(
    x: Any = None, maxlag: Any = None, addconst: Any = None, verbose: Any = None
) -> Any:
    if version.parse(statsmodels.__version__) < version.parse("0.15.0"):
        with warnings.catch_warnings():
            warnings.simplefilter(action="ignore", category=FutureWarning)
            return grangercausalitytests(x, maxlag, addconst, False)

    else:
        return grangercausalitytests(x, maxlag, addconst)


def _granger_causality_backend(
    data: pd.DataFrame, max_lag: Union[int, List[int]], add_constant: bool = True
) -> Dict[Any, Any]:
    assert len(data.columns) == 2, "`data` must have only two columns"
    assert (
        isinstance(max_lag, int)
        or isinstance(max_lag, list)
        and all(isinstance(l, int) for l in max_lag)
        and len(max_lag) > 0
    ), "`max_lag` must be an integer or a list of integers"
    assert isinstance(add_constant, bool), "`add_constant` must be a boolean"

    arguments: Dict[str, Any] = dict(
        x=data,
        maxlag=max_lag,
        addconst=add_constant,
    )

    return _statsmodels_compatibility_wrapper(**arguments)


def _get_tickers(
    tickers: Optional[List[str]] = None,
    cids: Optional[List[str]] = None,
    xcats: Optional[List[str]] = None,
) -> List[str]:
    """
    Simply returns the tickers if they are specified. If they are not specified, then
    the function forms the list of tickers from the `cids` and `xcats` such that the
    order of the formed tickers is preserved.

    :param <List[str]> tickers: A list of tickers.
    :param <Union[str, List[str]]> cids: One or two cids.
    :param <Union[str, List[str]]> xcats: One or two xcats.
    """

    if tickers is not None:
        return tickers
    else:
        if isinstance(cids, str):
            cids: List[str] = [cids]
        if isinstance(xcats, str):
            xcats: List[str] = [xcats]
        return [f"{c}_{x}" for c in cids for x in xcats]


def _type_checks(
    df: pd.DataFrame,
    tickers: Optional[List[str]],
    cids: Optional[List[str]],
    xcats: Optional[List[str]],
    max_lag: Union[int, List[int]],
    add_constant: bool,
    start: Optional[str],
    end: Optional[str],
    freq: str,
    agg: str,
    metric: str,
) -> bool:
    """
    Does type checks on the inputs to `granger_causality_test`.
    All inputs are checked for type and value errors.

    :params: See `granger_causality_test` for details.
    :return <bool>: True if all type checks pass.
    :raises <TypeError>: If any of the inputs are of the wrong type.
    :raises <ValueError>: If any of the input values are invalid.
    """

    if not isinstance(df, QuantamentalDataFrame):
        raise TypeError("df must be a standardized quantamental dataframe")

    if not isinstance(metric, str):
        raise TypeError("`metric` must be a string")

    err_msg: str = f"`metric` '{metric}' not found in `df`"
    if metric not in df.columns:
        raise ValueError(err_msg)

    if not isinstance(max_lag, (int, list)):
        raise TypeError("`max_lag` must be an integer or a list of integers")
    elif isinstance(max_lag, list) and not all(isinstance(l, int) for l in max_lag):
        raise TypeError("`max_lag` must be an integer or a list of integers")

    for dt, nm in zip([start, end], ["start", "end"]):
        if dt is not None and not is_valid_iso_date(dt):
            raise ValueError(f"{nm} must be a valid ISO date")

    if isinstance(cids, str):
        cids: List[str] = [cids]
    if isinstance(xcats, str):
        xcats: List[str] = [xcats]

    if bool(cids) ^ bool(xcats):
        raise ValueError("`cids` and `xcats` must be specified together")

    bcidxcats: bool = bool(cids) and bool(xcats)

    if bool(tickers) and (bcidxcats):
        raise ValueError(
            "`tickers` cannot be specified if `cids` & `xcats` are specified"
        )

    found_tickers: List[str] = (df["cid"] + "_" + df["xcat"]).unique().tolist()

    if bool(tickers):
        # check if there are only two
        if len(set(tickers)) != 2:
            raise ValueError("Only two tickers can be specified in `tickers`")

        if not all(isinstance(t, str) for t in tickers):
            raise TypeError("`tickers` must be a list of strings")

        if not set(tickers).issubset(set(found_tickers)):
            raise ValueError(
                "All tickers specified in `tickers` must be in `df`."
                f"Missing tickers: {set(tickers) - set(found_tickers)}"
            )

    else:
        assert bcidxcats, "Failed to resolve tickers"

    if bcidxcats:
        for lx, nm in zip([cids, xcats], ["cid", "xcat"]):
            if not (isinstance(lx, list) and all(isinstance(x, str) for x in lx)):
                raise TypeError(f"`{nm}` must be a list of strings")
            if not set(lx).issubset(set(df[nm])):
                raise ValueError(
                    f"All '{nm}s' in `{nm}` specified must be in `df`. "
                    f"Missing {nm}s: {set(lx) - set(df[nm])}."
                )

        tks: List[str] = [f"{c}_{x}" for c in cids for x in xcats]
        if not len(tks) == 2:
            raise ValueError(
                "The combination of `cids` & `xcats` must yield two tickers",
                f"Found {len(tks)} tickers: {tks}, ",
                f"from `cids` {cids} and `xcats` {xcats}",
            )
        if not set(tks).issubset(set(found_tickers)):
            raise ValueError(
                "All combinations of `cids` & `xcats` (i.e. tickers) specified must "
                "be in `df`."
                f"Missing tickers: {set(tks) - set(found_tickers)}"
            )

    if not isinstance(freq, str):
        raise TypeError("`freq` must be a string")

    if not isinstance(agg, str):
        raise TypeError("`agg` must be a string")

    if add_constant not in [True, False] or not isinstance(add_constant, bool):
        raise TypeError("`add_constant` must be a boolean")

    return True


def granger_causality_test(
    df: pd.DataFrame,
    tickers: Optional[List[str]] = None,
    cids: Optional[Union[str, List[str]]] = None,
    xcats: Optional[Union[str, List[str]]] = None,
    max_lag: Union[int, List[int]] = 4,
    add_constant: bool = False,
    freq: str = "M",
    agg: str = "mean",
    start: Optional[str] = None,
    end: Optional[str] = None,
    metric: str = "value",
) -> Dict[Any, Any]:
    """
    Run Granger Causality Test on a standardized quantamental dataframe.
    Since the Graner Causality Test is a pairwise test, only two tickers can be specified.
    Specify either `tickers` or `cids` & `xcats`. When specifying `cids` & `xcats`, the
    user may input one `cid` and two `xcats`; or two `cids` and one `xcat` to yield two
    tickers. The function forms the list of tickers from the `cids` and `xcats` such that
    the order of the formed tickers is preserved. The order of the tickers is important
    as the first ticker is the one that is tested to Granger cause the second ticker.
    The function tests whether the first ticker Granger causes the second ticker.

    :param <pd.DataFrame> df: A standardized quantamental dataframe.
    :param <List[str]> tickers: A list of tickers to run the test on. A maximum of two
        tickers can be specified.
    :param <Union[str, List[str]]> cids: One or two cids to run the test on. If two cids
        are specified, then only one xcat can be specified. If one cid is specified, then
        two xcats can be specified.
    :param <Union[str, List[str]]> xcats: One or two xcats to run the test on. If two
        xcats are specified, then only one cid can be specified. If one xcat is specified,
        then two cids can be specified.
    :param <Union[int, List[int]]> max_lag: If `max_lag` is an integer, then the function
        computes the test for all lags up to `max_lag`. If `max_lag` is a list of
        integers, then the function computes the test only for lags specified in the list.
    :param <bool> add_constant: Whether to add a constant to the regression.
    :param <str> freq: The frequency to downsample the data to. Must be one of "D", "W",
        "M", "Q", "A". Default is "M".
    :param <str> agg: The aggregation method to use when downsampling the data. Must be
        one of "mean" (default), "median", "min", "max", "first" or "last".
    :param <str> start: The start date of the data. Must be a valid ISO date. If not
        specified, the earliest date in `df` is used.
    :param <str> end: The end date of the data. Must be a valid ISO date. If not
        specified, the latest date in `df` is used.
    :param <str> metric: The metric to run the test on. Must be a column in `df`.
        Default is "value".

    :return <Dict[Any, Any]>: A dictionary containing the results of the Granger Causality
        Test. The keys are the lags and the values are the results of the test.
    :raises <TypeError>: If any of the inputs are of the wrong type.
    :raises <ValueError>: If any of the input values are invalid.
    """
    ## Check inputs

    _type_checks(
        df=df,
        tickers=tickers,
        cids=cids,
        xcats=xcats,
        max_lag=max_lag,
        add_constant=add_constant,
        start=start,
        end=end,
        freq=freq,
        agg=agg,
        metric=metric,
    )
    ## value checks for `freq` and `agg` are implicitly checked in downstream functions

    ## Copy df to prevent side effects
    df: QuantamentalDataFrame = df.copy()

    ## Construct tickers from the `cids` and `xcats` if `tickers` is not specified
    tickers: List[str] = _get_tickers(tickers=tickers, cids=cids, xcats=xcats)

    ## Reduce df
    df: pd.DataFrame = reduce_df_by_ticker(df=df, ticks=tickers, start=start, end=end)

    # Downsample df
    freq = freq.upper()
    agg = agg.lower()
    df = downsample_df_on_real_date(
        df=df, groupby_columns=["cid", "xcat"], freq=freq, agg=agg
    )

    # Pivot df
    df_wide: pd.DataFrame = qdf_to_ticker_df(df=df, value_column=metric)

    # there must only be two columns in df_wide
    assert len(df_wide.columns) == 2, "df_wide must have only two columns"

    logger.info(
        "Running Granger Causality Test: Testing whether %s Granger causes %s",
        df_wide.columns[0],
        df_wide.columns[1],
    )
    # NOTE: Since no NANs are allowed in the input data, we must drop them here
    # This may yield unexpected/unreliable results for tickers with large periods of
    # missing data

    # drop any rows with NANs
    df_wide = df_wide.dropna(how="any", axis=0)
    if df_wide.empty:
        raise ValueError(
            "The input data contains only NANs. "
            "Please check the input data for missing values or "
            "consider using a different downsampling frequency/date range."
        )

    gct: Dict[Any, Any] = _granger_causality_backend(
        data=df_wide,
        max_lag=max_lag,
    )

    return gct


if __name__ == "__main__":
    cids: List[str] = ["AUD"]
    xcats: List[str] = ["FX", "EQ"]

    df: pd.DataFrame = make_test_df(
        cids=cids,
        xcats=xcats,
    )

    gct: Dict[Any, Any] = granger_causality_test(
        df=df,
        cids=cids,
        xcats=xcats,
    )

    cids: List[str] = ["AUD", "CAD"]
    xcats: str = "FX"
    # tickers =  AUD_FX, CAD_FX
    df: pd.DataFrame = make_test_df(
        cids=cids,
        xcats=xcats,
    )

    gct: Dict[Any, Any] = granger_causality_test(
        df=df,
        tickers=["AUD_FX", "CAD_FX"],
    )

    print(gct)



import numpy as np
import pandas as pd
import datetime

from sklearn.linear_model import LinearRegression
from sklearn.base import RegressorMixin, BaseEstimator, clone
from macrosynergy.learning.predictors.bootstrap import BasePanelBootstrap

from abc import ABC, abstractmethod
from typing import Union, Optional

class BaseModifiedRegressor(BaseEstimator, RegressorMixin, BasePanelBootstrap, ABC):
    def __init__(
        self,
        model: RegressorMixin,
        method: str,
        error_offset: float = 1e-5,
        bootstrap_method: str = "panel",
        bootstrap_iters: int = 100,
        resample_ratio: Union[float, int] = 1,
        max_features: Optional[Union[str, int, float]] = None,
        analytic_method: Optional[str] = None,
    ):
        """
        Base class for linear regressors where coefficients are modified by estimated
        standard errors to account for statistical precision of the estimates.

        :param <RegressorMixin> model: The underlying linear model to be modified. This
            model must have `coef_` and `intercept_` attributes, in accordance with
            standard `scikit-learn` conventions.
        :param <str> method: The method used to modify the coefficients. Accepted values
            are "analytic" and "bootstrap".
        :param <float> error_offset: A small offset to add to the standard errors to
            prevent division by zero in the case of very small standard errors. Default
            value is 1e-5.
        :param <str> bootstrap_method: The bootstrap method used to modify the coefficients.
            Accepted values are "panel", "period", "cross", "cross_per_period"
            and "period_per_cross". Default value is "panel".
        :param <int> bootstrap_iters: The number of bootstrap iterations to perform in
            order to determine the standard errors of the model parameters under the bootstrap
            approach. Default value is 100.
        :param <Union[float, int]> resample_ratio: The ratio of resampling units comprised
            in each bootstrap dataset. This is a fraction of the quantity of the panel
            component to be resampled. Default value is 1.
        :param <Optional[Union[str, int, float]]> max_features: The number of features to
            consider in each bootstrap dataset. This can be used to increase the 
            variation of the bootstrap datasets. Default is None and currently not
            implemented.
        :param <Optional[str]> analytic_method: The analytic method used to calculate
            standard errors. Expressions for analytic standard errors are expected to be
            written within the method `adjust_analytical_se` and this parameter can be
            passed into `adjust_analyical_se` for an alternative analytic standard error
            estimate, for instance White's estimator. Default value is None.

        :return None
        """
        # Checks
        super().__init__(
            bootstrap_method=bootstrap_method,
            resample_ratio=resample_ratio,
            max_features=max_features,
        )

        self._check_init_params(
            model=model,
            method=method,
            error_offset=error_offset,
            bootstrap_iters=bootstrap_iters,
            analytic_method=analytic_method,
        )

        # Set attributes
        self.model = model
        self.method = method
        self.error_offset = error_offset
        self.bootstrap_iters = bootstrap_iters
        self.analytic_method = analytic_method

    def _check_init_params(
        self,
        model: RegressorMixin,
        method: str,
        error_offset: float,
        bootstrap_iters: int,
        analytic_method: Optional[str],
    ):
        """
        Method to check the validity of the initialization parameters of the class.

        :param <RegressorMixin> model: The underlying linear model to be modified. This
            model must have `coef_` and `intercept_` attributes once fit, in accordance
            with standard `scikit-learn` conventions.
        :param <str> method: The method used to modify the coefficients. Accepted values
            are "analytic" and "bootstrap".
        :param <float> error_offset: A small offset to add to the standard errors to
            prevent division by zero in the case of very small standard errors.
        :param <int> bootstrap_iters: The number of bootstrap iterations to perform in
            order to determine the standard errors of the model parameters under the
            bootstrap approach.
        :param <Optional[str]> analytic_method: The analytic method used to calculate
            standard errors. Expressions for analytic standard errors are expected to be
            written within the method `adjust_analytical_se` and this parameter can be
            passed into `adjust_analyical_se` for an alternative analytic standard error
            estimate, for instance White's estimator.

        :return None
        """
        # model
        if not isinstance(model, BaseEstimator):
            raise TypeError("model must be a valid `scikit-learn` estimator.")
        if not isinstance(model, RegressorMixin):
            raise TypeError("model must be a valid `scikit-learn` regressor.")
        
        # method
        if not isinstance(method, str):
            raise TypeError("method must be a string.")
        if method not in ["analytic", "bootstrap"]:
            raise ValueError("method must be either 'analytic' or 'bootstrap'.")
        
        # error_offset
        if not isinstance(error_offset, (float, int)):
            raise TypeError("error_offset must be a float or an integer.")
        if error_offset <= 0:
            raise ValueError("error_offset must be greater than 0.")
        
        # bootstrap_iters
        if method == "bootstrap":
            if not isinstance(bootstrap_iters, int):
                raise TypeError("bootstrap_iters must be an integer.")
            if bootstrap_iters <= 0:
                raise ValueError("bootstrap_iters must be a positive integer.")
            
        # analytic_method
        if method == "analytic":
            if analytic_method is not None:
                if not isinstance(analytic_method, str):
                    raise TypeError("analytic_method must be a string.")
                
    def fit(
        self,
        X: pd.DataFrame,
        y: Union[pd.DataFrame, pd.Series],
    ):
        """
        Fit method to fit the underlying linear model, as passed into the constructor,
        and subsequently modify coefficients based on estimated standard errors to produce
        a trading signal that accounts for the statistical precision of the factor weights.

        :param <pd.DataFrame> X: Pandas dataframe of input features.
        :param <Union[pd.DataFrame, pd.Series]> y: Pandas series or dataframe of targets
            associated with each sample in X.

        :return <BaseModifiedRegressor>
        """
        # Checks
        self._check_fit_params(X=X, y=y)

        # Fit
        self.model.fit(X, y)

        if not hasattr(self.model, "coef_"):
            raise AttributeError("The underlying model must have a `coef_` attribute.")
        if not hasattr(self.model, "intercept_"):
            raise AttributeError(
                "The underlying model must have an `intercept_` attribute."
            )
        
        # Modify coefficients
        if self.method == "analytic":
            self.intercept_, self.coef_ = self.adjust_analytical_se(
                self.model,
                X,
                y,
                self.analytic_method,
            )
        elif self.method == "bootstrap":
            # clone the model to avoid modifying the original model
            model = clone(self.model)
            self.intercept_, self.coef_ = self.adjust_bootstrap_se(
                model,
                X,
                y,
            )

        return self
    
    def predict(
        self,
        X: pd.DataFrame,
    ):
        """
        Method to predict the target variable using the underlying linear model.

        :param <pd.DataFrame> X: Pandas dataframe of input features.
        """
        # Checks
        if not isinstance(X, pd.DataFrame):
            raise TypeError(
                "Input feature matrix must be a pandas dataframe. "
                "If used as part of an sklearn pipeline, ensure that previous steps "
                "return a pandas dataframe."
            )
        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("X must be multi-indexed.")
        
        return self.model.predict(X)

    def create_signal(
        self,
        X: pd.DataFrame,
    ):
        """
        Method to create a signal based on adjusting the underlying factor model weights
        by their standard errors.

        :param <pd.DataFrame> X: Pandas dataframe of input features.
        """
        # Checks
        if not isinstance(X, pd.DataFrame):
            raise TypeError(
                "Input feature matrix must be a pandas dataframe. "
                "If used as part of an sklearn pipeline, ensure that previous steps "
                "return a pandas dataframe."
            )
        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("X must be multi-indexed.")

        return np.dot(X, self.coef_) + self.intercept_
    
    def adjust_bootstrap_se(
        self,
        model: RegressorMixin,
        X: pd.DataFrame,
        y: Union[pd.DataFrame, pd.Series],
    ):
        # Create storage for bootstrap coefficients and intercepts
        bootstrap_coefs = np.zeros((self.bootstrap_iters, X.shape[1]))
        bootstrap_intercepts = np.zeros(self.bootstrap_iters)

        # Bootstrap loop
        for i in range(self.bootstrap_iters):
            X_resampled, y_resampled = self.create_bootstrap_dataset(X, y)
            model.fit(X_resampled, y_resampled)
            bootstrap_coefs[i] = model.coef_
            bootstrap_intercepts[i] = model.intercept_

        # Calculate standard errors
        coef_se = np.std(bootstrap_coefs, axis=0, ddof=0)
        intercept_se = np.std(bootstrap_intercepts, ddof=0)

        # Adjust the coefficients and intercepts by the standard errors
        coef = self.model.coef_ / (coef_se + self.error_offset)
        intercept = self.model.intercept_ / (intercept_se + self.error_offset)

        return intercept, coef
    
    def adjust_analytical_se(
        self,
        model: RegressorMixin,
        X: pd.DataFrame,
        y: Union[pd.DataFrame, pd.Series],
        analytic_method: Optional[str],
    ):
        raise NotImplementedError(
            "Analytical standard error adjustments are not available for most models."
            "This function must be implemented in a subclass of BaseModifiedRegressor "
            "if known standard error expressions are available."
        )
    
    def _check_fit_params(
        self,
        X: pd.DataFrame,
        y: Union[pd.DataFrame, pd.Series],
    ):
        """
        Method to check the validity of the fit parameters of the class.

        :param <pd.DataFrame> X: Pandas dataframe of input features.
        :param <Union[pd.DataFrame, pd.Series]> y: Pandas series or dataframe of targets
        """
        # Checks
        if not isinstance(X, pd.DataFrame):
            raise TypeError(
                "Input feature matrix must be a pandas dataframe. "
                "If used as part of an sklearn pipeline, ensure that previous steps "
                "return a pandas dataframe."
            )
        if not (isinstance(y, pd.Series) or isinstance(y, pd.DataFrame)):
            raise TypeError(
                "Target vector must be a pandas series or dataframe. "
                "If used as part of an sklearn pipeline, ensure that previous steps "
                "return a pandas series or dataframe."
            )
        if isinstance(y, pd.DataFrame) and y.shape[1] != 1:
            raise ValueError(
                "The target dataframe must have only one column. If used as part of "
                "an sklearn pipeline, ensure that previous steps return a pandas "
                "series or dataframe."
            )

        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("X must be multi-indexed.")
        if not isinstance(y.index, pd.MultiIndex):
            raise ValueError("y must be multi-indexed.")
        if not isinstance(X.index.get_level_values(1)[0], datetime.date):
            raise TypeError("The inner index of X must be datetime.date.")
        if not isinstance(y.index.get_level_values(1)[0], datetime.date):
            raise TypeError("The inner index of y must be datetime.date.")
        if not X.index.equals(y.index):
            raise ValueError(
                "The indices of the input dataframe X and the output dataframe y don't "
                "match."
            )
        
class ModifiedLinearRegression(BaseModifiedRegressor):
    def __init__(
        self,
        method: str,
        fit_intercept: bool = True,
        positive: bool = False,
        error_offset: float = 1e-2,
        bootstrap_method: str = "panel",
        bootstrap_iters: int = 1000,
        resample_ratio: Union[float, int] = 1,
        analytic_method: Optional[str] = None,
    ):
        """
        Custom class to train an OLS linear regression model with coefficients modified
        by estimated standard errors to account for statistical precision of the estimates.

        :param <str> method: The method used to modify the coefficients. Accepted values
            are "analytic" and "bootstrap".
        :param <bool> fit_intercept: Whether to fit an intercept term in the model.
            Default is True.
        :param <bool> positive: Whether to constrain the coefficients to be positive.
            Default is False.
        :param <float> error_offset: A small offset to add to the standard errors to
            prevent division by zero in the case of very small standard errors. Default
            value is 1e-2.
        :param <str> bootstrap_method: The bootstrap method used to modify the coefficients.
            Accepted values are "panel", "period", "cross", "cross_per_period"
            and "period_per_cross". Default value is "panel".
        :param <int> bootstrap_iters: The number of bootstrap iterations to perform in
            order to determine the standard errors of the model parameters under the bootstrap
            approach. Default value is 1000.
        :param <Union[float, int]> resample_ratio: The ratio of resampling units comprised
            in each bootstrap dataset. This is a fraction of the quantity of the panel
            component to be resampled. Default value is 1.
        :param <Optional[str]> analytic_method: The analytic method used to calculate
            standard errors. Expressions for analytic standard errors are expected to be
            written within the method `adjust_analytical_se` and this parameter can be
            passed into `adjust_analyical_se` for an alternative analytic standard error
            estimate, for instance White's estimator. Default value is None.

        :return None
        """
        self.fit_intercept = fit_intercept
        self.positive = positive

        super().__init__(
            model=LinearRegression(
                fit_intercept=self.fit_intercept, positive=self.positive
            ),
            method=method,
            error_offset=error_offset,
            bootstrap_method=bootstrap_method,
            bootstrap_iters=bootstrap_iters,
            resample_ratio=resample_ratio,
            analytic_method=analytic_method,
        )

    def adjust_analytical_se(
        self,
        model: RegressorMixin,
        X: pd.DataFrame,
        y: Union[pd.DataFrame, pd.Series],
        analytic_method: Optional[str],
    ):
        """
        Method to adjust the coefficients of the linear model by an analytical
        standard error expression. The default is to use the standard error estimate
        obtained through assuming multivariate normality of the model errors as well as
        heteroskedasticity and zero mean. If `analytic_method` is "White", the White
        estimator is used to estimate the standard errors.

        :param <RegressorMixin> model: The underlying linear model to be modified. This
            model must have `coef_` and `intercept_` attributes, in accordance with
            standard `scikit-learn` conventions.
        :param <pd.DataFrame> X: Pandas dataframe of input features.
        :param <Union[pd.DataFrame, pd.Series]> y: Pandas series or dataframe of targets
            associated with each sample in X.
        :param <Optional[str]> analytic_method: The analytic method used to calculate
            standard errors. If None, the default method is used. Currently, the only
            alternative we offer is White's estimator, which requires "White" to be
            specified. Default value is None.

        :return <float>, <np.ndarray>: The adjusted intercept and coefficients.
        """
        # Checks
        if analytic_method is not None:
            if not isinstance(analytic_method, str):
                raise TypeError("analytic_method must be a string.")
            if analytic_method not in ["White"]:
                raise ValueError("analytic_method must be 'White'.")

        if self.fit_intercept:
            X_new = np.column_stack((np.ones(len(X)), X.values))
        else:
            X_new = X.values

        # Calculate the standard errors
        predictions = model.predict(X)
        residuals = (y - predictions).to_numpy()
        XtX_inv = np.linalg.inv(X_new.T @ X_new)
        if analytic_method is None:
            se = np.sqrt(
                np.diag(
                    XtX_inv
                    * np.sum(np.square(residuals))
                    / (X.shape[0] - X.shape[1])
                )
            )

        elif analytic_method == "White":
            # Implement HC3
            leverages = np.sum((X_new @ XtX_inv) * X_new, axis=1)
            weights = 1 / (1 - leverages) ** 2
            residuals_squared = np.square(residuals)
            weighted_residuals_squared = weights * residuals_squared
            Omega = X_new.T * weighted_residuals_squared @ X_new
            cov_matrix = XtX_inv @ Omega @ XtX_inv
            se = np.sqrt(np.diag(cov_matrix))

        else:
            raise NotImplementedError(
                "Currently, only the standard and White standard errors are implemented"
            )

        if self.fit_intercept:
            coef_se = se[1:]
            intercept_se = se[0]
        else:
            coef_se = se
            intercept_se = 0

        # Adjust the coefficients and intercepts by the standard errors
        coef = model.coef_ / (coef_se + self.error_offset)
        intercept = model.intercept_ / (intercept_se + self.error_offset)

        return intercept, coef

    def set_params(self, **params):
        super().set_params(**params)
        if "fit_intercept" in params or "positive" in params:
            # Re-initialize the LinearRegression instance with updated parameters
            self.model = LinearRegression(
                fit_intercept=self.fit_intercept, positive=self.positive
            )

        return self


if __name__ == "__main__":
    from macrosynergy.management.simulate import make_qdf
    import macrosynergy.management as msm
    from macrosynergy.learning import SignalOptimizer, ExpandingKFoldPanelSplit

    from sklearn.linear_model import LinearRegression
    from sklearn.model_selection import GridSearchCV
    from sklearn.metrics import make_scorer, r2_score

    # Randomly generate an unbalanced panel dataset, multi-indexed by cross-section and
    # real_date

    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]
    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2002-01-01", "2020-12-31", 0, 1]
    df_cids.loc["CAD"] = ["2003-01-01", "2020-12-31", 0, 1]
    df_cids.loc["GBP"] = ["2000-01-01", "2020-12-31", 0, 1]
    df_cids.loc["USD"] = ["2000-01-01", "2020-12-31", 0, 1]

    df_xcats = pd.DataFrame(index=xcats, columns=cols)
    df_xcats.loc["XR"] = ["2000-01-01", "2020-12-31", 0, 1, 0, 3]
    df_xcats.loc["CRY"] = ["2000-01-01", "2020-12-31", 0, 1, 0, 0]
    df_xcats.loc["GROWTH"] = ["2000-01-01", "2020-12-31", 0, 1, -0.9, 0]
    df_xcats.loc["INFL"] = ["2000-01-01", "2020-12-31", 0, 1, 0.8, 0]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    dfd["grading"] = np.ones(dfd.shape[0])
    black = {"GBP": ["2009-01-01", "2012-06-30"], "CAD": ["2018-01-01", "2100-01-01"]}
    dfd = msm.reduce_df(df=dfd, cids=cids, xcats=xcats, blacklist=black)

    dfd = dfd.pivot(index=["cid", "real_date"], columns="xcat", values="value")
    X = dfd.drop(columns=["XR"])
    y = dfd["XR"]

    # First instantiate the BaseModifiedRegressor
    obj = BaseModifiedRegressor(model=LinearRegression(), method="bootstrap")

    # Demonstrate ModifiedLinearRegression usage
    method_pairs = [
        ("analytic", "panel", None),
        ("analytic", "panel", "White"),
        ("bootstrap", "panel", None),
        ("bootstrap", "period", None),
        ("bootstrap", "cross", None),
        ("bootstrap", "cross_per_period", None),
        ("bootstrap", "period_per_cross", None),
    ]
    for method in method_pairs:
        model = ModifiedLinearRegression(
            method=method[0],
            bootstrap_method=method[1],
            bootstrap_iters=100,
            resample_ratio=0.75,
            analytic_method=method[2],
        )
        # Fit the model
        model.fit(X, y)
        print("----")
        print("Modified OLS method:", method)
        print("Modified OLS intercept:", model.intercept_)
        print("Modified OLS coefficients:", model.coef_)
        print("Modified OLS predictions:", model.predict(X))
        print("Modified OLS signal:", model.create_signal(X))
        print("----")

        # Grid search
        cv = GridSearchCV(
            estimator=model,
            param_grid={
                "fit_intercept": [True, False],
                "positive": [True, False],
            },
            cv=5,
            n_jobs=-1,
        )
        cv.fit(X, y)
        print("----")
        print("Modified OLS grid search results:")
        print(pd.DataFrame(cv.cv_results_))
        print("----")

        # Try with signal optimizer
        inner_splitter = ExpandingKFoldPanelSplit(n_splits=5)
        so = SignalOptimizer(
            inner_splitter=inner_splitter,
            X=X,
            y=y,
        )
        so.calculate_predictions(
            name=f"ModifiedOLS_{method[0]}_{method[1]}",
            models={
                "mlr": model,
            },
            metric=make_scorer(r2_score, greater_is_better=True),
            hparam_grid={
                "mlr": {
                    "fit_intercept": [True, False],
                    "positive": [True, False],
                },
            },
            hparam_type="grid",
            test_size=21 * 12 * 2,
            n_jobs = -1
        )
        so.models_heatmap(f"ModifiedOLS_{method[0]}_{method[1]}")
        so.coefs_stackedbarplot(f"ModifiedOLS_{method[0]}_{method[1]}")



from . import predictors

from .panel_time_series_split import (
    ExpandingKFoldPanelSplit,
    RollingKFoldPanelSplit,
    ExpandingIncrementPanelSplit,
    ExpandingFrequencyPanelSplit,
    BasePanelSplit,
)
from .cv_tools import panel_cv_scores
from .transformers import (
    LassoSelector,
    MapSelector,
    ENetSelector,
    ZnScoreAverager,
    PanelMinMaxScaler,
    PanelStandardScaler,
    FeatureAverager,
)
from .metrics import (
    neg_mean_abs_corr,
    panel_significance_probability,
    sharpe_ratio,
    sortino_ratio,
    regression_accuracy,
    regression_balanced_accuracy,
)
from .signal_optimizer import SignalOptimizer

from .predictors import (
    NaivePredictor,
    BaseWeightedRegressor,
    WeightedLinearRegression,
    SignWeightedLinearRegression,
    TimeWeightedLinearRegression,
    WeightedLADRegressor,
    SignWeightedLADRegressor,
    TimeWeightedLADRegressor,
    LADRegressor,
    BaseRegressionSystem,
    LADRegressionSystem,
    RidgeRegressionSystem,
    LinearRegressionSystem,
    CorrelationVolatilitySystem,
    ModifiedLinearRegression,
    BaseModifiedRegressor,
)

from .beta_estimator import BetaEstimator

__all__ = [
    # panel_time_series_split
    "ExpandingKFoldPanelSplit",
    "RollingKFoldPanelSplit",
    "ExpandingIncrementPanelSplit",
    "ExpandingFrequencyPanelSplit",
    "BasePanelSplit",
    # cv_tools
    "panel_cv_scores",
    # transformers
    "FeatureAverager",
    "LassoSelector",
    "MapSelector",
    "ENetSelector",
    "PanelMinMaxScaler",
    "PanelStandardScaler",
    "ZnScoreAverager",
    # metrics
    "neg_mean_abs_corr",
    "panel_significance_probability",
    "sharpe_ratio",
    "sortino_ratio",
    "regression_accuracy",
    "regression_balanced_accuracy",
    # signal_optimizer
    "SignalOptimizer",
    # predictors
    "LADRegressor",
    "NaivePredictor",
    "SignWeightedLADRegressor",
    "BaseWeightedRegressor",
    "TimeWeightedLADRegressor",
    "SignWeightedLinearRegression",
    "TimeWeightedLinearRegression",
    "ModifiedLinearRegression",
    "BaseModifiedRegressor",
    # market beta estimation
    "BetaEstimator",
    "WeightedLinearRegression",
    "WeightedLADRegressor",
    # regression system
    "BaseRegressionSystem",
    "LADRegressionSystem",
    "RidgeRegressionSystem",
    "LinearRegressionSystem",
    "CorrelationVolatilitySystem",
]
